package com.paxos.testutils

import com.google.protobuf.ProtocolMessageEnum
import com.paxos.eltnam.DateRange
import com.paxos.eltnam.toJodaDateTime
import io.kotest.property.Arb
import io.kotest.property.RandomSource
import io.kotest.property.arbitrary.arbitrary
import io.kotest.property.arbitrary.bind
import io.kotest.property.arbitrary.bool
import io.kotest.property.arbitrary.choice
import io.kotest.property.arbitrary.constant
import io.kotest.property.arbitrary.enum
import io.kotest.property.arbitrary.filter
import io.kotest.property.arbitrary.localDate
import io.kotest.property.arbitrary.localDateTime
import io.kotest.property.arbitrary.long
import io.kotest.property.arbitrary.map
import io.kotest.property.arbitrary.of
import io.kotest.property.arbitrary.set
import io.kotest.property.arbitrary.string
import io.kotest.property.arbitrary.stringPattern
import io.kotest.property.arbitrary.uuid
import org.joda.time.DateTime
import java.lang.Integer.max
import java.math.BigDecimal
import java.time.LocalDate
import java.util.UUID

fun nonZeroBigDecimal(rs: RandomSource): BigDecimal {
    val beforeDecimal = Arb.long(0, 1000000000).sample(rs).value
    val afterDecimal = Arb.long(0, 1000000000).samples(rs).filter {
        if (beforeDecimal == 0L) {
            it.value != 0L
        } else {
            true
        }
    }.first().value
    return BigDecimal("$beforeDecimal.$afterDecimal")
}

fun nonNegativeBigDecimalGen(): Arb<BigDecimal> =
    arbitrary(listOf(BigDecimal.ZERO, BigDecimal.ONE)) { nonZeroBigDecimal(it) }

fun negativeBigDecimalGen(): Arb<BigDecimal> = arbitrary(listOf(BigDecimal("-1"))) { -nonZeroBigDecimal(it) }
fun bigDecimalGen(): Arb<BigDecimal> = arbitrary(listOf(BigDecimal.ZERO, BigDecimal.ONE, BigDecimal.ONE.negate())) {
    if (Arb.bool().sample(it).value) {
        -nonZeroBigDecimal(it)
    } else {
        nonZeroBigDecimal(it)
    }
}

fun nonUUIDStringGen(): Arb<String> = arbitrary(listOf("")) {
    Arb.string().sample(it).value
}

fun genNonEmptyString(
    rs: RandomSource = RandomSource.default(),
    minSize: Int = 1,
    maxSize: Int = 100,
): String =
    genStringIf(rs, minSize = max(1, minSize), maxSize) { str -> str.isNotBlank() }

fun Arb.Companion.nonEmptyString(): Arb<String> =
    Arb.string(minSize = 1).filter { it.isNotBlank() }

fun Arb.Companion.jodaDateTime(): Arb<DateTime> =
    Arb.localDateTime().map { it.toJodaDateTime() }

fun genDateTime(rs: RandomSource = RandomSource.default()): DateTime =
    Arb.jodaDateTime().sample(rs).value

fun genDate(rs: RandomSource = RandomSource.default()): LocalDate =
    Arb.localDate().sample(rs).value

fun genUuid(rs: RandomSource = RandomSource.default()): UUID =
    Arb.uuid().sample(rs).value

fun maybeGenNonEmptyString(rs: RandomSource = RandomSource.default()): String? =
    maybeGen(rs, ::genNonEmptyString)

fun maybeGenDateTime(rs: RandomSource = RandomSource.default()): DateTime? =
    maybeGen(rs, ::genDateTime)

fun maybeGenUuid(rs: RandomSource = RandomSource.default()): UUID? =
    maybeGen(rs, ::genUuid)

fun <T> generateIf(rs: RandomSource, predicate: () -> Boolean, action: (RandomSource) -> T): T? =
    if (predicate()) action(rs) else null

fun <T> maybeGen(rs: RandomSource, action: (RandomSource) -> T): T? =
    if (Arb.bool().sample(rs).value) action(rs) else null

fun <T> arbNullable(arbFun: () -> Arb<T>): Arb<T?> =
    Arb.choice(Arb.constant(null), arbFun())

fun <T> arbNullIf(predicate: () -> Boolean, arbFun: () -> Arb<T>): Arb<T?> =
    if (predicate()) Arb.constant(null) else arbFun()

fun <T> arbConstantIfNotNull(value: T?, arbFun: () -> Arb<T>): Arb<T> =
    value?.let { Arb.constant(value) } ?: arbFun()

inline fun <reified T : Enum<T>> generateEnum(rs: RandomSource = RandomSource.default()): T =
    Arb.enum<T>().sample(rs).value

fun arbNonEmptyString(maxSize: Int = 100) = Arb.string(1, maxSize)
fun arbAlphanumeric() = Arb.stringPattern("[a-zA-Z0-9]+")

fun arbDateRangeNullableStart() = Arb.bind(arbNullable { Arb.localDate() }, Arb.localDate()) { startDate, endDate ->
    DateRange(startDate, endDate)
}

fun arbDateRangeNullableEnd() = Arb.bind(Arb.localDate(), arbNullable { Arb.localDate() }) { startDate, endDate ->
    DateRange(startDate, endDate)
}

fun arbDateRange() = Arb.choice(arbDateRangeNullableStart(), arbDateRangeNullableEnd())

inline fun <reified T> arbProtoEnum() where T : ProtocolMessageEnum, T : Enum<T> =
    Arb.of(
        enumValues<T>().filter {
            try {
                it.number
                true
            } catch (e: java.lang.IllegalArgumentException) {
                false
            }
        }
    )

fun genUniqueNonEmptyStrings(
    numStrings: Int,
    maxStringSize: Int = 100,
    rs: RandomSource = RandomSource.default(),
): Set<String> =
    Arb.set(Arb.string(minSize = 1, maxSize = maxStringSize), IntRange(numStrings, numStrings)).sample(rs).value

fun genStringIf(
    rs: RandomSource = RandomSource.default(),
    minSize: Int = 0,
    maxSize: Int = 100,
    condition: (String) -> Boolean,
): String =
    Arb.string(minSize, maxSize).samples(rs).filter { sampleStr -> condition(sampleStr.value) }.first().value
